/**********************************************************************
  sim floppy (Pseudo floppyhCu)
  $Id: dev_floppy.c,v 1.3 2003/07/06 01:37:02 tane Exp $
**********************************************************************/
#include "common.h"
#include "ppcsim.h"
#include "processor.h"
#include "funcs.h"
#include "memory.h"
#include "op_exec.h"
#include "sim_io.h"

#define FDC03f0  0x800003f0
#define FDC03f1  0x800003f1
#define FDC03f2  0x800003f2
#define FDC03f3  0x800003f3
#define FDC03f4  0x800003f4
#define FDC03f5  0x800003f5
#define FDC03f6  0x800003f6
#define FDC03f7  0x800003f7

#define FDC_STAT_NOTREADY	0
#define FDC_STAT_READY		1

#define FDC_CMD_NOP             0x00
#define FDC_CMD_SENSE_INT       0x08
#define FDC_CMD_CONFIGURE       0x13
#define FDC_CMD_SPECIFY         0x03
#define FDC_CMD_RECALIBRATE     0x07
#define FDC_CMD_READ            0x06
#define FDC_CMD_READ_TRACK      0x02
#define FDC_CMD_READ_ID         0x0A
#define FDC_CMD_DUMP_REG        0x0E
#define FDC_CMD_SEEK            0x0F

#define FDC_CMD_SENSE_INT_LEN   0x01
#define FDC_CMD_CONFIGURE_LEN   0x04
#define FDC_CMD_SPECIFY_LEN     0x03
#define FDC_CMD_RECALIBRATE_LEN 0x02
#define FDC_CMD_READ_LEN        0x09
#define FDC_CMD_READ_TRACK_LEN  0x09
#define FDC_CMD_READ_ID_LEN     0x02
#define FDC_CMD_DUMP_REG_LEN    0x01
#define FDC_CMD_SEEK_LEN        0x03

#define FDC_RES_SENSE_INT_LEN   0x02
#define FDC_RES_CONFIGURE_LEN   0x04
#define FDC_RES_SPECIFY_LEN     0x03
#define FDC_RES_RECALIBRATE_LEN 0x02
#define FDC_RES_READ_LEN        0x09
#define FDC_RES_READ_TRACK_LEN  0x09
#define FDC_RES_READ_ID_LEN     0x02
#define FDC_RES_DUMP_REG_LEN    0x01
#define FDC_RES_SEEK_LEN        0x02

#define FDC_STATE_CMDINPUT  0
#define FDC_STATE_RESOUTPUT 1


struct FDC_struct{
  unsigned int reset ;
  unsigned int state ;
  unsigned int drive ;
  unsigned char dor ;
  unsigned char sra ;
  unsigned char msr ;
} ;
struct FDD_struct{
  unsigned int cmd_type ;
  unsigned int cmd_stat ;
  unsigned int cmd_len ;
  unsigned int res_stat ;
  unsigned int res_len ;
  unsigned char cmd_buf[16] ;
  unsigned char res_buf[16] ;
  
  unsigned int mt ;
  unsigned int mfm ;
  unsigned int skip ;
  unsigned int head ;
} ;

struct FDC_struct FDC ;
struct FDD_struct FDD[4] ;

static int fdc_interrupt_delay ;

int dev_floppy( unsigned int port , struct MACHINE_RESOURCE *mres )
{
  if( __access_fdc==1 ){
    SYSMSG("FDC(pc:%08x,ad:%08x,%s) = %02x %02x %02x %02x %02x %02x %02x %02x\n",
	   __access_fdc_pc , __access_fdc_adrs ,
	   (__access_fdc_is_ld==1 ? "ld" : "st") ,
	   mres->mem[seg(0x800003f0)][lad(0x800003f0)],
	   mres->mem[seg(0x800003f1)][lad(0x800003f1)],
	   mres->mem[seg(0x800003f2)][lad(0x800003f2)],
	   mres->mem[seg(0x800003f3)][lad(0x800003f3)],
	   mres->mem[seg(0x800003f4)][lad(0x800003f4)],
	   mres->mem[seg(0x800003f5)][lad(0x800003f5)],
	   mres->mem[seg(0x800003f6)][lad(0x800003f6)],
	   mres->mem[seg(0x800003f7)][lad(0x800003f7)]) ;
  }

  /******************************/
  /******************************/
  if( __access_fdc==1 ){
	if( __access_fdc_is_ld==1 ){//read
	  switch( __access_fdc_adrs & 0x7 ){
	  case 0x0 : //FDC_SRA
		mres->mem[seg(FDC03f2)][lad(FDC03f2)]=FDC.sra ;
		break ;
	  case 0x1 : //FDC_SRB
		break ;
	  case 0x2 : //FDC_DOR
		mres->mem[seg(FDC03f2)][lad(FDC03f2)]=FDC.dor ;
		break ;
	  case 0x3 : //FDC_TDR
		break ;
	  case 0x4 : //FDC_DSR|FDC_MSR
		if( FDC.state==FDC_STATE_CMDINPUT ){
		  FDC.msr &= ~(0x40) ;
		}else{
		  FDC.msr |= 0x40 ;
		}
		mres->mem[seg(FDC03f4)][lad(FDC03f4)]=FDC.msr ;
		if( (FDC.dor & 0x08)!=0 ){
		  dev_pic_set_interrupt( PIC_IR06_INT ) ;
		}
		break ;
	  case 0x5 : //FDC_FIFO
		if( FDC.state==FDC_STATE_RESOUTPUT ){
		  if( FDD[FDC.drive].res_stat==0 ){
			switch( FDD[FDC.drive].cmd_type ){
			case FDC_CMD_SENSE_INT :
			  FDD[FDC.drive].res_len=FDC_RES_SENSE_INT_LEN ;
			  break ;
			case FDC_CMD_CONFIGURE :
			  break ;
			default :
			  FDD[FDC.drive].cmd_type=FDC_CMD_NOP ;
			}
		  }
		  mres->mem[seg(FDC03f5)][lad(FDC03f5)]=FDD[FDC.drive].res_buf[FDD[FDC.drive].res_stat] ;
		  if( FDD[FDC.drive].cmd_type!=FDC_CMD_NOP ){
			FDD[FDC.drive].res_stat++ ;
			if( FDD[FDC.drive].res_stat==FDD[FDC.drive].res_len ){
			  FDC.state=FDC_STATE_CMDINPUT ;
			  FDD[FDC.drive].cmd_type=FDC_CMD_NOP ;
			  FDD[FDC.drive].cmd_stat=0 ;
			  FDD[FDC.drive].cmd_len =0 ;
			  FDD[FDC.drive].res_stat=0 ;
			  FDD[FDC.drive].res_len =0 ;
			}
		  }
		}
		break ;
	  case 0x6 : //FDC_DIR
		break ;
	  case 0x7 : //FDC_CCR
		break ;
	  default :
		break ;
	  }
	}else{ // write
	  switch( __access_fdc_adrs & 0x7 ){
	  case 0x0 : //FDC_SRA
		break ;
	  case 0x1 : //FDC_SRB
		break ;
	  case 0x2 : //FDC_DOR
		FDC.dor=mres->mem[seg(FDC03f2)][lad(FDC03f2)] ;
		if( FDC.dor==0x00 ){ // FDC reset
		  FDD[FDC.drive].cmd_stat = 0x00 ;
		  FDC.msr      = 0x80 ;
		}
		FDC.sra |= 0x80 ;
		FDC.drive=FDC.dor & 0x03 ;
		mres->mem[seg(FDC03f4)][lad(FDC03f4)]=FDC.msr ;
		break ;
	  case 0x3 : //FDC_TDR
		break ;
	  case 0x4 : //FDC_DSR|FDC_MSR
		break ;
	  case 0x5 : //FDC_FIFO
		if( FDC.state==FDC_STATE_CMDINPUT ){
		  unsigned char cmd ;
		  cmd=mres->mem[seg(FDC03f5)][lad(FDC03f5)] ;
		  if( FDD[FDC.drive].cmd_stat == 0 ){
			FDD[FDC.drive].cmd_type = cmd & 0x1f ;
			switch( FDD[FDC.drive].cmd_type ){
			case FDC_CMD_SENSE_INT :
			  FDD[FDC.drive].cmd_len=FDC_CMD_SENSE_INT_LEN ;
			  break ;
			default :
			  FDD[FDC.drive].cmd_type=FDC_CMD_NOP ;
			  break ;
			}
		  }
		  if( FDD[FDC.drive].cmd_type!=FDC_CMD_NOP ){
			FDD[FDC.drive].cmd_buf[FDD[FDC.drive].cmd_stat]=cmd ;
			FDD[FDC.drive].cmd_stat++ ;
			if( FDD[FDC.drive].cmd_stat==FDD[FDC.drive].cmd_len ){
			  FDC.state=FDC_STATE_RESOUTPUT ;
			  FDD[FDC.drive].cmd_stat=0 ;
			  FDD[FDC.drive].cmd_len =0 ;
			  FDD[FDC.drive].res_stat=0 ;
			  FDD[FDC.drive].res_len =0 ;
			}
		  }
		}
		break ;
	  case 0x6 : //FDC_DIR
		break ;
	  case 0x7 : //FDC_CCR
		break ;
	  default :
		break ;
	  }
	}
  }
  return(0) ;
}


int dev_floppy_init( struct MACHINE_RESOURCE *mres )
{
  FDC.reset   = FDC_STAT_READY ;
  FDC.drive   = 0 ;
  FDC.state   = FDC_STATE_CMDINPUT ;
  FDC.dor     = 0x00 ;
  FDC.sra     = 0x00 ;
  FDC.msr     = 0x80 ;
  FDD[FDC.drive].cmd_stat= 0 ;
  FDD[FDC.drive].cmd_type= 0 ;
  fdc_interrupt_delay=0 ;
  return( NORMAL_EXIT ) ;
}

/*

 R}h͂AFDC.MSRreadyݒ肷

*/
