/************************************************************************
  sim PIC (Pseudo荞݃Rg[)
  $Id: dev_pic.c,v 1.3 2003/07/06 01:37:02 tane Exp $
************************************************************************/
#include "common.h"
#include "ppcsim.h"
#include "processor.h"
#include "funcs.h"
#include "memory.h"
#include "op_exec.h"
#include "sim_io.h"

#define WAIT_FOR_ICW1_IN   0
#define WAIT_FOR_ICW2_IN   1
#define WAIT_FOR_ICW3_IN   2
#define WAIT_FOR_ICW4_IN   3

static int pic_interrupt_delay ;

int dev_pic( unsigned int port , struct MACHINE_RESOURCE *mres )
{
  int id ;
  int dev ;
  int data ;
  
  data=mres->mem[seg(port)][lad(port)] ;
  
  id= port & 0x00000001 ;
  dev=0 ;
  if( (port & 0x00000f0)==0xa0 ){
    dev=1 ;
  }
  
  switch( id ){
  case 0x0 :
    switch( data & 0x00000018 ){
    case 0x00000010 :
    case 0x00000018 :
      /* ICW1 */
      PIC_REG[dev].icw1=data ;
      PIC_REG[dev].state=WAIT_FOR_ICW2_IN ;
      break ;
    case 0x00000000 :
      PIC_REG[dev].ocw2=data ;
      if( (PIC_REG[dev].ocw2 & 0xe0)!=0x40 ){
		PIC_REG[dev].IRR &= ~PIC_REG[dev].ISR ;
		PIC_REG[dev].ISR = 0 ;
		dev_pic_prio_int() ;
      }
      break ;
    case 0x00000008 :
      PIC_REG[dev].ocw3=data ;
      break ;
    default :
      break ;
    }
    if(       (PIC_REG[dev].ocw3 & 0x03)==0x02 ){
      mres->mem[seg(PIC_MASTER_PORT0)][lad(PIC_MASTER_PORT0)]=PIC_REG[dev].IRR ;
    }else if( (PIC_REG[dev].ocw3 & 0x03)==0x01 ){
      mres->mem[seg(PIC_MASTER_PORT0)][lad(PIC_MASTER_PORT0)]=PIC_REG[dev].ISR ;
    }
    break ;
    
  case 0x1 :
    if(       PIC_REG[dev].state==WAIT_FOR_ICW1_IN ){
      PIC_REG[dev].IMR=data ; /** same set ocw1 **/
    }else if( PIC_REG[dev].state==WAIT_FOR_ICW2_IN ){
      PIC_REG[dev].icw2=data ;
      PIC_REG[dev].state=WAIT_FOR_ICW3_IN ;
    }else if( PIC_REG[dev].state==WAIT_FOR_ICW3_IN ){
      PIC_REG[dev].icw3=data ;
      if( (PIC_REG[dev].icw1 & 1)==0 ){
		PIC_REG[dev].state=WAIT_FOR_ICW1_IN ;
      }else{
		PIC_REG[dev].state=WAIT_FOR_ICW4_IN ;
      }
    }else if( PIC_REG[dev].state==WAIT_FOR_ICW4_IN ){
      PIC_REG[dev].icw4=data ;
      PIC_REG[dev].state=WAIT_FOR_ICW1_IN ;
    }
    mres->mem[seg(PIC_MASTER_PORT0)][lad(port)]=PIC_REG[dev].IMR ; /* store IMR */
    
    break ;
    
  default :
    return(-1) ;
  }
  
  DEBUGMSG("PIC%d : IMR=%02x , ISR=%02x , IRR=%02x\n",
		   dev,PIC_REG[dev].IMR,PIC_REG[dev].ISR,PIC_REG[dev].IRR) ;
  return(0) ;
}


int dev_pic_set_interrupt(int dev_int)
{
  PIC_REG[1].IRR |= dev_int & 0x0000ff00 ;
  PIC_REG[0].IRR |= dev_int & 0x000000ff ;
  if( PIC_REG[1].IRR!=0 ){
    PIC_REG[0].IRR |=PIC_IR02_INT ;
  }
  pic_interrupt_delay=10000 ;
  //dev_pic_prio_int() ;
  return(0) ;
}


int dev_pic_prio_int()
{
  unsigned int ir ;
  unsigned int prio_flag ;
  int i ;

  if( pic_interrupt_delay < 0 ){
	if( PIC_REG[0].ISR==0 && PIC_REG[1].ISR==0 ){
	  if( PIC_REG[0].IRR!=0 || PIC_REG[1].IRR!=0 ){
		ir=((PIC_REG[1].IRR & ~PIC_REG[1].IMR)<<8) |
		  ((PIC_REG[0].IRR & ~PIC_REG[0].IMR)) ;
		for( i=0,prio_flag=1 ; i<16 ; i++,prio_flag<<=1 ){
		  if( (ir & prio_flag)!=0 ){
			PIC_REG[0].ISR=  (ir & prio_flag) & 0x00ff ;
			PIC_REG[1].ISR= ((ir & prio_flag) & 0xff00)>>8 ;
			EXTERNAL_INT=1 ;
			break ;
		  }
		}
	  }
	}
  }else{
	pic_interrupt_delay-- ;
  }
  return(0) ;
}


int dev_pic_init( struct MACHINE_RESOURCE *mres )
{
  PIC_REG[0].state=WAIT_FOR_ICW1_IN ;
  PIC_REG[0].icw1 =0x11 ;
  PIC_REG[0].icw2 =0x08 ;
  PIC_REG[0].icw3 =0x04 ;
  PIC_REG[0].icw4 =0x01 ;
  PIC_REG[0].IMR  =0x00 ;
  PIC_REG[0].ocw2 =0x00 ;
  PIC_REG[0].ocw3 =0x00 ;
  PIC_REG[0].IRR  =0x00 ;
  PIC_REG[0].ISR  =0x00 ;
  
  PIC_REG[1].state=WAIT_FOR_ICW1_IN ;
  PIC_REG[1].icw1 =0x11 ;
  PIC_REG[1].icw2 =0x70 ;
  PIC_REG[1].icw3 =0x02 ;
  PIC_REG[1].icw4 =0x01 ;
  PIC_REG[1].IMR  =0x00 ;
  PIC_REG[1].ocw2 =0x00 ;
  PIC_REG[1].ocw3 =0x00 ;
  PIC_REG[1].IRR  =0x00 ;
  PIC_REG[1].ISR  =0x00 ;
  
  mres->mem[seg(PIC_MASTER_PORT0)][lad(PIC_MASTER_PORT0)]=0 ;
  mres->mem[seg(PIC_MASTER_PORT1)][lad(PIC_MASTER_PORT1)]=PIC_REG[0].IMR ;
  mres->mem[seg(PIC_SLAVE_PORT0) ][lad(PIC_SLAVE_PORT0) ]=0 ;
  mres->mem[seg(PIC_SLAVE_PORT1) ][lad(PIC_SLAVE_PORT1) ]=PIC_REG[1].IMR ;
  
  return(0) ;
}

