/**********************************************************************
  O
  $Id: exception.c,v 1.5 2002/11/03 03:57:35 tane Exp $
**********************************************************************/
#include "common.h"
#include "ppcsim.h"
#include "processor.h"
#include "funcs.h"
#include "memory.h"
#include "op_exec.h"

int exception( struct MACHINE_RESOURCE *mres , int code )
{
int ret_code ;
 
 ret_code=0 ;
 if( (code & OPFETCH_FLAG)==OPFETCH_FLAG ){
   switch( (code & ~OPFETCH_FLAG) ){
   case PC_BOUNDARY_ERROR :  /** ppcsim [O **/
     SYSMSG("pc=%08x : PC Boundary Error.\n",mres->pc) ;
     ret_code=-1 ;
     break ;
   case MEM_AREA_OVER :
     if( excp_stat.e0400.stat==ENABLE ){
	*mres->SPR[26] = mres->pc                    ; /** SRR0 **/
        *mres->SPR[27] = ( *mres->SPR[27] & 0x87c0008c ) | \
	                 ( *mres->MSR & 0x0000ff73 ) ; /** SRR1 **/
	*mres->MSR = *mres->MSR & ~(0x0004ef30) ;
	mres->pc=excp_stat.e0400.pc ;
	ret_code=1 ;
     }else{
       SYSMSG("pc=%08x : Instruction fetch area is over.\n",mres->pc) ;
       ret_code=-1 ;
     }
     break ;
   case NON_PHYMEM :
     SYSMSG("pc=%08x : Physical memory is not alloced.\n",mres->pc) ;
     ret_code=-1 ;
     break ;
   default :
     SYSMSG("pc=%08x : PPCSIM internal error.\n",mres->pc) ;
     ret_code=-1 ;
   }
 }else{
   switch( code ){
   case UNKNOWN_INSTRUCTION :
     if( excp_stat.e0700.stat==ENABLE ){
	mres->pc+=4 ;
	*mres->SPR[26] = mres->pc                ; /** SRR0 **/
	*mres->SPR[27] = *mres->MSR & 0x87e00000 ; /** SRR1 **/
	mres->pc=excp_stat.e0700.pc ;
     }else{
       SYSMSG("pc=%08x : Unknown Instruction.\n",mres->pc) ;
       ret_code=-1 ;
     }
     break ;
   case DIVIDED_BY_ZERO : /** ppcsim [O **/
     SYSMSG("pc=%08x : Divided by Zero.\n",mres->pc) ;
     ret_code=-1 ;
     break ;
   case INVALID_BO :
     if( excp_stat.e0700.stat==ENABLE ){
	mres->pc+=4 ;
	*mres->SPR[26] = mres->pc                ; /** SRR0 **/
	*mres->SPR[27] = *mres->MSR & 0x87e00000 ; /** SRR1 **/
	mres->pc=excp_stat.e0700.pc ;
     }else{
       SYSMSG("pc=%08x : BO is invalid.\n",mres->pc) ;
       ret_code=-1 ;
     }
     break ;
   case EXCEPTION_PROGRAM :
     if( excp_stat.e0700.stat==ENABLE ){
	mres->pc+=4 ;
	*mres->SPR[26] = mres->pc                ; /** SRR0 **/
	*mres->SPR[27] = *mres->MSR & 0x87e00000 ; /** SRR1 **/
	mres->pc=excp_stat.e0700.pc ;
     }else{
       SYSMSG("pc=%08x : Detect trap.\n",mres->pc) ;
       ret_code=-1 ;
     }
     break ;
   case INSTFORM_IS_INVALID :
     if( excp_stat.e0700.stat==ENABLE ){
	mres->pc+=4 ;
	*mres->SPR[26] = mres->pc                ; /** SRR0 **/
	*mres->SPR[27] = *mres->MSR & 0x87e00000 ; /** SRR1 **/
	mres->pc=excp_stat.e0700.pc ;
     }else{
       SYSMSG("pc=%08x : Instruction format is invalid.\n",mres->pc) ;
       ret_code=-1 ;
     }
     break ;
   case UNKNOWN_SYSCALL_NO :
     SYSMSG("pc=%08x : Unknown SystemCall Number(%d).\n",mres->pc,mres->GPR[0].rl) ;
     ret_code=-1 ;
     break ;
   case NOT_SUPPORT_SC_EMU :
     SYSMSG("pc=%08x : Not include SystemCall Emulation.\n",mres->pc) ;
     ret_code=-1 ;
     break ;
   case NOT_IMPLEMENTED_OP :
     SYSMSG("pc=%08x : Not implemented operation.\n",mres->pc) ;
     ret_code=-1 ;
     break ;
   case MEM_AREA_OVER :
     if( excp_stat.e0300.stat==ENABLE ){
	*mres->SPR[18] = mres->inst                  ; /** DSISR **/
	*mres->SPR[19] = __ea_for_exception          ; /** DAR   **/
	mres->pc+=4 ;
	*mres->SPR[26] = mres->pc                    ; /** SRR0 **/
        *mres->SPR[27] = ( *mres->SPR[27] & 0x87c0008c ) | \
	                 ( *mres->MSR & 0x0000ff73 ) ; /** SRR1 **/
	*mres->MSR = *mres->MSR & ~(0x0004ef30) ;
	mres->pc=excp_stat.e0300.pc ;
     }else{
       SYSMSG("pc=%08x : Memory access area is over. ea=%08x pa=%08x\n",
	      mres->pc,__ea_for_exception,__pa_for_exception ) ;
       ret_code=-1 ;
     }
     break ;
   case SYSCALLED_EXIT :
     SYSMSG("pc=%08x : exit() was called. Simulation is stopped.\n",mres->pc) ;
     ret_code=-1 ;
     break ;
   case EXCEPTION_SYSCALL :
     if( excp_stat.e0c00.stat==ENABLE ){
       *mres->SPR[26] = mres->pc                    ; /** SRR0 **/
       *mres->SPR[27] = ( *mres->SPR[27] & 0x87c008fc ) | \
	                ( *mres->MSR & 0x0000f703 ) ; /** SRR1 **/
       *mres->MSR = *mres->MSR & ~(0x0004ef30) ;
       mres->pc=excp_stat.e0c00.pc ;
       DEBUGMSG("pc=%08x : Catch SYSTEM-CALL exception.\n",mres->pc) ;
     }
     break ;
   case EXCEPTION_EE :
     if( excp_stat.e0500.stat==ENABLE ){
       *mres->SPR[26] = mres->pc                    ; /** SRR0 **/
       *mres->SPR[27] = ( *mres->SPR[27] & 0x87c008fc ) | \
	                ( *mres->MSR & 0x0000f703 ) ; /** SRR1 **/
       *mres->MSR = *mres->MSR & ~(0x0004ef30) ;
       mres->pc=excp_stat.e0500.pc ;
       DEBUGMSG("pc=%08x : Catch External Interrupt.\n",mres->pc) ;
     }else{
       if( excp_stat.e0500.stat==DISABLE ){
	 SYSMSG("pc=%08x : Arrival External Interrupt.\n",mres->pc) ;
	 ret_code=-1 ;
       }else{
	 SYSMSG("pc=%08x : Arrival External Interrupt.but losted\n",mres->pc) ;
       }
     }
     break ;
   case EXCEPTION_DEC :
     if( excp_stat.e0900.stat==ENABLE ){
       *mres->SPR[26] = mres->pc                    ; /** SRR0 **/
       *mres->SPR[27] = ( *mres->SPR[27] & 0x87c008fc ) | \
	                ( *mres->MSR & 0x0000f703 ) ; /** SRR1 **/
       *mres->MSR = *mres->MSR & ~(0x0004ef30) ;
       mres->pc=excp_stat.e0900.pc ;
       DEBUGMSG("pc=%08x : Catch decrement exception.\n",mres->pc) ;
     }else{
       if( excp_stat.e0900.stat==DISABLE ){
	 SYSMSG("pc=%08x : Arrival decrement exception.\n",mres->pc) ;
	 ret_code=-1 ;
       }else{
	 SYSMSG("pc=%08x : Arrival decrement exception. but losted.\n",mres->pc) ;
	 print_regs( mres, (PRINT_CRS|PRINT_GPRS) ) ;
       }
     }
     break ;
   case EXCEPTION_SE :
     if( excp_stat.e0d00.stat==ENABLE ){
       *mres->SPR[26] = mres->pc                    ; /** SRR0 **/
       *mres->SPR[27] = ( *mres->SPR[27] & 0x87c008fc ) | \
	                ( *mres->MSR & 0x0000f703 ) ; /** SRR1 **/
       *mres->MSR = *mres->MSR & ~(0x0004ef30) ;
       mres->pc=excp_stat.e0d00.pc ;
       DEBUGMSG("pc=%08x : Catch single step exception.\n",mres->pc) ;
     }else{
       if( excp_stat.e0d00.stat==DISABLE ){
	 SYSMSG("pc=%08x : Arrival single step exception.\n",mres->pc) ;
	 ret_code=-1 ;
       }else{
	 SYSMSG("pc=%08x : Arrival single step exception. but losted.\n",mres->pc) ;
       }
     }
     break ;
   case NON_PHYMEM :
     SYSMSG("pc=%08x : Physical memory is not alloced.\n",mres->pc) ;
     ret_code=-1 ;
     break ;
   default :
     SYSMSG("pc=%08x : PPCSIM internal error.\n",mres->pc) ;
     ret_code=-1 ;
   }
 }
 return( ret_code ) ;
}
  
