/**********************************************************************
  [v[uR}h
  $Id: prob_cmd.c,v 1.4 2003/07/06 01:37:04 tane Exp $
**********************************************************************/
#include "common.h"
#include "ppcsim.h"
#include "processor.h"
#include "funcs.h"
#include "memory.h"
#include "op_exec.h"

int  prob_parse_opt(struct GETOPT_ARG *args) ;
void prob_print_help() ;

int prob_cmd( struct GETOPT_ARG *args , int mode )
{
  int i ;
  int ret ;
  
  ret=prob_parse_opt(args) ;
  if( ret<0 ) return(NORMAL_EXIT) ;

  if( XGETOPT_REFARGP(args,1)==NULL ){
    if( MemprobeBuf.num==0 ){
    }else{
      SYSMSG("---- memory probe ----\n") ;
      for( i=0 ; i<MemprobeBuf.num ; i++ ){
	SYSMSG("0x%08x size=0x%08x",
	       MemprobeBuf.adrs[i],
	       MemprobeBuf.size[i]) ;
	if( MemprobeBuf.label[i]!=NULL ){
	  SYSMSG(" : %s",MemprobeBuf.label[i]) ;
	}
	SYSMSG("\n") ;
      }
    }
    return(NORMAL_EXIT) ;
  }

  if( MemprobeBuf.num <= MEMPROBE_BUFSIZE ){
    if( XGETOPT_REFARGP(args,3)!=NULL ){
      MemprobeBuf.label[MemprobeBuf.num]=malloc(strlen(XGETOPT_REFARGP(args,3))+1) ;
      if( MemprobeBuf.label[MemprobeBuf.num]==NULL ){
	return( MEM_ALLOC_ERROR ) ;
      }
      strcpy(MemprobeBuf.label[MemprobeBuf.num],XGETOPT_REFARGP(args,3)) ;
    }

    MemprobeBuf.adrs[MemprobeBuf.num]=exatoi(XGETOPT_REFARGP(args,1)) ;
    if( XGETOPT_REFARGP(args,2)==NULL ){
      MemprobeBuf.size[MemprobeBuf.num]=1 ;
    }else{
      int n ;
      n=exatoi(XGETOPT_REFARGP(args,2)) ;
      if( n<1 ){
	SYSMSG("Size is Invalid. (%d)\n",n) ;
	return( NORMAL_EXIT ) ;
      }
      MemprobeBuf.size[MemprobeBuf.num]=n ;
    }
    MemprobeBuf.num++ ;
  }else{
    SYSMSG("probe point is full.\n") ;
  }

  return(NORMAL_EXIT) ;
}


int prob_parse_opt(struct GETOPT_ARG *args)
{
  static struct GETOPT_LIST getopt_list[]={
    {"help",               'h',  NONARG, 'h'},
    { NULL,               '\0',       0,'\0'} /*** terminator ***/
  } ;
  int ret ;
  
  while(-1){
    ret=xgetopt(args, getopt_list) ;
    if( ret==XGETOPT_PARSE_COMPLETE ) break ;
    switch(ret){
    case 'h' :
      prob_print_help() ;
      return(-1) ;
      break ;
    case XGETOPT_INVALID_OPTION :
      SYSERR("Invalid option %s\n",xgetopt_err) ;
      return(-1) ;
      break ;
    case XGETOPT_ARG_NOT_FOUND :
      SYSERR("Illegal arg for option %s\n",xgetopt_err) ;
      return(-1) ;
      break ;
    case XGETOPT_NOT_OPTION :
      break ;
    default :
      SYSERR("xgetopt internal error\n") ;
      return(-1) ;
    }
  }
  return(0) ;
}


void prob_print_help()
{
  SYSERR(
	 "probe memory\n"
	 "Usage: prob [adrs] [size] [label]\n"
	 "  adrs  : Memory address.\n"
	 "  size  : Probing memory area size.\n"
	 "  label : Probing point name.\n"
	 " Options\n"
	 "  -h or --help : Print this help.\n"
	 "\n"
	 "     goR}hɂV~[VsAprobR}hŎw肵\n"
	 "    AhXTCỸԂɑ΂ă[h/XgȂs߂\n"
	 "    i[AhX\\܂B\n"
	 "     TCYȗꍇsize=1ƂĐݒ肳܂B\n"
	 "\n"
	 "     ȗꍇ͌ݐݒ肳Ă郁v[uAhẌꗗ\n"
	 "    \\܂B\n"
	 ) ;
}
